import arrow
import os
import json
from datetime import datetime
from abc import ABC

NO_PERSON = 'no_person'


def pst_to_dt(time):
    local = datetime.strptime(
        time, "%Y-%m-%d %H:%M:%S")
    return arrow.get(local, 'US/Pacific').to('utc')

class CustomerInfo(ABC):
    def __init__(self):
        ci_file = os.getenv('CUSTOMER_INFO_FILE', '/etc/customer_info.json')
        try:
            with open(ci_file, encoding='utf-8') as json_file:
                self.customer_data = json.load(json_file)
        except Exception as err:
            raise Exception("failed to load customer info %s : %s" %
                            (ci_file, err))

    def list_persons(self):
        persons = self.customer_data.keys()
        return persons

    def list_domains(self, person=None):
        domains = []
        for p, data  in self.customer_data.items():
            if person and not p == person:
                continue
            if 'domains' in data:
                try:
                    domains.append([d['domain'] for d in data['domains']])
                except Exception as e:
                    domains.append([d for d in data['domains']])
        return domains

    def get_persons(self):
        persons = []
        # putting all these into an array uses a ton of memory
        for person, data in self.customer_data.items():
            persons.append(self.get_person(person))
        return persons

    def person_by_user(self, user):
        for person, data in self.customer_data.items():
            if ('users' in data):
                users = []
                try:
                    users = [u['user'] for u in data['users']]
                except Exception as e:
                    users = [u for u in data['users']]

                if user in users:
                    return Person(person, data)
        return Person(NO_PERSON)

    def person_by_domain(self, domain):
        for person, data in self.customer_data.items():
            if ('domains' in data):
                domains = [d['domain'] for d in data['domains'] if d['domain'] == domain]
                if domains:
                    return Person(person, data)
        return Person(NO_PERSON)

    def get_person(self, person):
        person_data = {}
        if person in self.customer_data:
            person_data = self.customer_data[person]
        return Person(person, person_data)

    def get_domain(self, domain):
        for person, data in self.customer_data.items():
            if ('domains' in data):
                for domain_data in data['domains']:
                    if domain_data['domain'] == domain:
                        return Domain(domain_data['domain'], domain_data)
        # else return an empty domain object
        return Domain(domain, {})

class Person(ABC):
    def __init__(self, person, data={}):
        default_dict = {'person': person}
        if not data:
            default_dict = {'person': NO_PERSON}
        self.__dict__ = {**default_dict, **data}

    def get_domains(self):
        domains = []
        if hasattr(self, 'domains'):
            domains = [Domain(d['domain'], d) for d in self.domains]
        return domains

    def seconds_since_joined(self, stamp=None):
        if not hasattr(self, 'when_joined'):
            return
        when_joined = pst_to_dt(self.when_joined)
        if not stamp:
            stamp = arrow.utcnow()
        return (stamp - when_joined).total_seconds()

    def should_send_events(self):
        if hasattr(self, 'events'):
            if self.events:
                return True
        return False

    def no_person(self):
        if self.person == NO_PERSON:
            return True


class Domain(ABC):
    def __init__(self, domain, data={}):
        default_dict = {'domain': domain}
        self.__dict__ = {**default_dict, **data}

    def no_domain(self):
        # TODO if there is no created we assume the object didn't load
        if not hasattr(self, 'created'):
            return True

    def person(self):
        return CustomerInfo().person_by_domain(self.domain)
