import time
import os
import sys
import traceback
import psutil
import statistics
from ..core.version import get_version
from ..core.scheduler import Scheduler
import plow.core.scrapers.apache
import plow.core.scrapers.auth
import plow.core.poller.dhwp
import plow.core.poller.sa
from cement import Controller, ex, shell
from cement.utils.version import get_version_banner
from plow.core.helpers import MetricState, EventState, Timer


VERSION_BANNER = """
Plow is a tool to scrape and export metrics %s
%s
""" % (get_version(), get_version_banner())


class Base(Controller):
    class Meta:
        label = 'base'

        # text displayed at the top of --help output
        description = 'Plow is a tool to scrape and export metrics'

        # text displayed at the bottom of --help output
        epilog = 'Usage: plow command1 --foo bar'

        # controller level arguments. ex: 'plow --version'
        arguments = [
            # add a version banner
            (['-v', '--version'],
             {'action': 'version',
                'version': VERSION_BANNER}),
        ]

    def _default(self):
        """Default action if no sub-command is passed."""

        self.app.args.print_help()

    @ex(
        help='start jobs',
        arguments=[
            (['-f', '--jobs_file'],
             {'help': 'jobs yaml file',
                'action': 'store',
                'default': '/etc/plow/jobs.yml',
                'dest': 'jobs_file'}),
            (['--metrics_path'],
             {'help': 'file(s) or path(s) with metrics',
                'action': 'store',
                'default': '/var/lib/plow',
                'dest': 'metrics_path'}),
        ],
    )
    def worker(self):
        threads = {}
        # wait between while loop
        sleep_wait = 5
        process = psutil.Process(os.getpid())
        memory_stats = []
        # get 5 min average of memory usage
        memory_stat_avg_minutes = 5
        metrics_state = MetricState(
            name='plow-worker',
            path=self.app.config.get('plow', 'metrics_path')
        )

        while True:
            if 'jobs' in threads:
                if not threads['jobs'].is_alive():
                    self.app.log.error("jobs thread died")
                    os._exit(1)

            if 'jobs' not in threads:
                self.app.log.info("starting jobs")
                scheduler = Scheduler(
                    logger=self.app.log, app_config=self.app.config['plow'])
                # TODO if a spawn fails.. exit the whole program
                try:
                    threads['jobs'] = shell.spawn(scheduler.start, kwargs={
                                                  'jobs_file': self.app.pargs.jobs_file}, thread=True)
                except Exception as e:
                    self.app.log.error("failed to spawn job worker")
                    os._exit(1)

            memory_stats.append(process.memory_info().rss)
            if len(memory_stats) >= ((60 / sleep_wait) * memory_stat_avg_minutes):
                memory_mean = statistics.mean(memory_stats)
                self.app.log.debug("5 min memory average %s" % memory_mean)
                memory_stats = []
                metrics_state.gauge(
                    'plow_memory_mean_bytes',
                    tags=['process'],
                    tag_values=['worker'],
                    value=memory_mean
                )

            metrics_state.dump(debug=self.app.config['plow'].get('debug'))
            time.sleep(sleep_wait)
        os._exit(1)

    @ex(
        help='events',
        arguments=[
            (['-t', '--type'],
             {'help': 'type',
                'action': 'store',
                'required': True,
                'dest': 'metric_type'}),
            (['-m', '--metric'],
             {'help': 'type',
                'action': 'store',
                'required': True,
                'dest': 'metric'}),
        ],
    )
    def events(self):
        state_name = '%s-%s' % (self.app.pargs.metric_type,
                                self.app.pargs.metric)

        state_path = self.app.config.get('plow', 'state_path')
        if os.path.exists(os.path.join(state_path, 'events-' + state_name + '.json')):
            event_state = EventState(
                name='events-' + state_name,
                path=state_path,
                rotate=60
            )
            for key in event_state.read():
                print(key)
        else:
            self.app.log.fatal("events do not exist for state %s" % state_name)

    @ex(
        help='metrics',
        arguments=[
            (['-t', '--type'],
             {'help': 'type',
                'action': 'store',
                'required': True,
                'dest': 'metric_type'}),
            (['-m', '--metric'],
             {'help': 'type',
                'action': 'store',
                'required': True,
                'dest': 'metric'}),
        ],
    )
    def metrics(self):
        from prometheus_exporter.datastore import MetricsDataStoreManager
        metric_name = '%s-%s' % (self.app.pargs.metric_type,
                                 self.app.pargs.metric)
        metrics_path = self.app.config.get('plow', 'metrics_path')
        metrics_file = os.path.join(metrics_path, metric_name + '.json')
        if os.path.exists(metrics_file):
            manager = MetricsDataStoreManager.deserialize(metrics_file)
            print(manager.dump_prometheus_metrics().decode("utf-8"))
        else:
            self.app.log.fatal(
                "metrics do not exist at %s" % metrics_file)

    @ex(
        help='scrape -t apache -l /path/to/mega.log',
        arguments=[
            (['-l', '--logs'],
             {'help': 'logs to scrape',
                'action': 'store',
                'required': True,
                'dest': 'logs'}),
            (['-t', '--type'],
             {'help': 'type of scrape',
                'action': 'store',
                'required': True,
                'dest': 'scrape_type'}),
            (['--loop'],
             {'help': 'keep looping this scrape',
                'action': 'store_true',
                'required': False,
                'default': False,
                'dest': 'loop'}),
        ],
    )
    def scrape(self):
        fast_scrape = self.app.config.get('plow', 'fast_scrape')
        self.app.log.info("scrape %s logs %s" %
                          (self.app.pargs.scrape_type, self.app.pargs.logs))

        while True:
            scrape_name = self.app.pargs.scrape_type
            scrape_obj = self.app.pargs.scrape_type.capitalize()
            try:
                cls = getattr(plow.core.scrapers, scrape_name)
                s = getattr(cls, scrape_obj)(logger=self.app.log,
                                             app_config=self.app.config['plow'])
                s.do_run(logs=self.app.pargs.logs, fast_scrape=fast_scrape)
            except Exception as err:
                self.app.log.error("failed to run scraper %s: %s" %
                                   (scrape_obj, err))
                self.app.log.error('%s' % __file__, exc_info=True)

            if not self.app.pargs.loop:
                break
            else:
                time.sleep(5)

    @ex(
        help='poll -t dhwp',
        arguments=[
            (['-t', '--type'],
             {'help': 'type of poll',
                'action': 'store',
                'required': True,
                'dest': 'poll_type'}),
            (['-f', '--force'],
             {'help': 'force polling',
                'action': 'store_true',
                'required': False,
                'dest': 'force'}),
        ],
    )
    def poll(self):
        self.app.config.set('plow', 'force', self.app.pargs.force)
        name = self.app.pargs.poll_type
        obj = name.capitalize()
        try:
            cls = getattr(plow.core.poller, name)
            s = getattr(cls, obj)(logger=self.app.log,
                                  app_config=self.app.config['plow'])
            s.do_run()
        except Exception as err:
            self.app.log.error("failed to run poller %s: %s" %
                               (obj, err))
            traceback.print_exc(file=sys.stdout)

    def _scrape_class(self, scrape_type):
        scrape_obj = scrape_type.capitalize()
        s = None
        try:
            cls = getattr(plow.core.scrapers, scrape_type)
            s = getattr(cls, scrape_obj)(logger=self.app.log,
                                         app_config=self.app.config['plow'])
        except Exception as e:
            self.app.log.error("failed to get scrape class %s" % scrape_type)
        return s
