import sys
import socket
import os
import pwd
import time
import statistics
from abc import ABC
from customer_info.ci import CustomerInfo
from event_api_client.core import EventAPI, Event, StateAPI
from event_api_client.vector import TopicHttpClient
from subprocess import Popen, PIPE, TimeoutExpired, SubprocessError
from plow.core.helpers import MetricState, Timer, load_too_high
from multiprocessing import cpu_count


class Poll(ABC):
    def __init__(self, logger, app_config=None, max_load=cpu_count() * 2):
        self.poll_type = type(self).__name__.lower()
        self.log = logger
        self.customer_info = CustomerInfo()
        self.max_load = max_load if max_load > 20 else 20
        if hasattr(self.log, '_meta'):
            self.log._meta.namespace = "plow:poll:%s" % (self.poll_type)

        self.app_config = app_config
        self.event_api = EventAPI(
            api_url=app_config['event_api_url'], api_token=app_config['event_api_token'])

    def add_state(self, **kwargs):
        with TopicHttpClient() as session:
            self.log.debug('putting message on vector topic: %s' % self.format_state(**kwargs))
            return session.create_message_for_topic(
                'rem_state_diff',
                self.format_state(**kwargs),
                user_agent='plow')

    def disable_load_check(self):
        return False

    def do_command(self, command, machine=None, timeout=10, user=None):
        if not self.disable_load_check():
            if load_too_high(max_load=self.max_load) and not self.app_config['force']:
                self.log.error("skipping poll command, load too high")
                return True

        if user:
            command = "sudo -u %s -n %s" % (user, command)

        if machine:
            myhostname = socket.gethostname()
            # TODO find a better way then just hostnames to do this
            if machine != myhostname:
                if 'lxchost' in myhostname:
                    command = "lxc exec %s -- %s" % (machine, command)
                elif 'vshost' in myhostname:
                    command = "vserver %s exec %s" % (machine, command)
                else:
                    self.log.error(
                        "not sure how to run the command on %s" % machine)
                    return

        self.log.debug("run command as user %s: %s" % (user, command))

        # TODO when everything is on python3.9, no more sudo is needed
        # will use Popen(user='blah') instead
        with Timer() as t:
            try:
                proc = Popen((command), stdout=PIPE,
                             stderr=PIPE, shell=True)
                (out, err) = proc.communicate(timeout=timeout)
                proc.wait()
            except TimeoutExpired as t:
                self.log.error("command timed out %s" % t)
                return
            except SubprocessError as e:
                self.log.error("command failed : %" % e)
                return

        code = proc.returncode
        if code > 0:
            self.log.error("command failed : %s" % command)
            return
        else:
            self.log.debug("command finished. took %s seconds" % t.seconds)
            self.timer_metrics['commands'].append(t.seconds)
            out = out.decode("utf-8")
            err = err.decode("utf-8")
            return self.process_results(out)

    def do_run(self, **kwargs):
        if load_too_high(max_load=self.max_load) and not self.app_config['force']:
            self.log.error("skipping poll, load too high")
            return True

        self.log.info("started poll %s" % self.poll_type)
        self.timer_metrics = {'commands': []}
        self.metrics_state = MetricState(
            name='poll-' + self.poll_type,
            path=self.app_config['metrics_path']
        )
        with Timer() as t:
            self.run(**kwargs)
        self.log.info("completed poll %s in %s seconds" % (
            self.poll_type, t.seconds))

        if self.timer_metrics['commands']:
            self.metrics_state.gauge(
                'plow_poll_command_mean_time',
                tags=['poll'],
                tag_values=[self.poll_type],
                value=statistics.mean(self.timer_metrics['commands'])
            )

        self.metrics_state.gauge(
            'plow_poll_total_time',
            tags=['poll'],
            tag_values=[self.poll_type],
            value=t.seconds
        )

        self.metrics_state.dump(debug=self.app_config.get('debug'))

    def run(self, **kwargs):
        pass

    def process_results(self, result):
        pass
